local socket = require("socket")
local connected = false
local udp = socket.udp()
local timeReceivedData = 0
local tryingToConnect = false

local broadcast = socket.udp()
local broadcasting = false

local karmaSendRate = 1			-- how many seconds for cooldown
local karmaSendCooldown = 0

function setupBroadcast()
	broadcast:settimeout(0)
	broadcast:setoption('broadcast', true)
	--udp:setoption('dontroute', true)
	broadcast:setsockname("*", B_PORT)
	broadcasting = true
end

function receiveNewWinner(playerName)
	roundRunning = false
	winnerName = playerName
end

function resumeGame()
	roundRunning = true
end

function receiveBroadcast()
	if not broadcasting then return end
	local b_pack, ip, port = broadcast:receivefrom()

	while b_pack do
		if b_pack == B_SERVER_MESSAGE then
			debug("Found a server with ip " .. ip .. " !")
			servers[ip] = true
		end
		b_pack, ip, port = broadcast:receivefrom()
	end
end

function sendBroadcast()
	if not broadcasting then return end
	debug("Send broadcast")
	local success, e_msg = broadcast:sendto(B_CLIENT_MESSAGE, B_ADDRESS, B_PORT)
	if not success then
		debug("Broadcast failed!")
		newStatusMsg("Failed to look for servers. Do you have one open? Then type 'localhost'.")
		debug(e_msg)
	end
end

function coolDownKarmaSend(dt)
	if karmaSendCooldown < karmaSendRate then
		karmaSendCooldown = karmaSendCooldown + dt
	end
end

function connect(ip, playerName)	-- connects to server and sends
	if #playerName == 0 then
		resetMenu()
		newStatusMsg("Empty playername!")
	elseif #ip == 0 then
		resetMenu()
		newStatusMsg("Empty IP-Address!")
	end
	
	tryingToConnect = true
	timeReceivedData = socket.gettime()
	udp:settimeout(0)
	udp:setpeername(ip, G_PORT)		-- G_PORT defined in common.lua
	
	udp:send("NEW:" .. playerName)
	debug ("connected...")
	newStatusMsg("Connecting...")
end

function chooseServer(ip)
	connect(ip, getMyPlayerName())
end

function checkForTimeOut()
	if tryingToConnect and (socket.gettime() - timeReceivedData > 10) then			-- timeout after 30 seconds
		disconnect()
		newStatusMsg("Timeout. Disconnected from server.")
		resetMenu()
		tryingToConnect = false
	end
end


function createNewPlayer(playerName)
	users[playerName] = {karma = 100, y = 0, sinOffset1 = math.random(3), sinOffset2 = math.random(3)}		-- if no user with name present, create him
end

function sendKarma(params)		-- direction is either "+" or "-"
	local playerName = params[1]
	local direction = params[2]
	if connected and karmaSendCooldown >= karmaSendRate then
		karmaSendCooldown = 0
		if direction == "-" or direction == "+" then
			udp:send("GIVE:" .. playerName .. "," .. direction)
		end
	end
end

local myLastKarma = 0

function handleNewData(data)
	debug("Received: " .. data)
	
	tryingToConnect = false		--remove as soon as ping is implemented
	timeReceivedData = socket.gettime()
	connected = true
	if getStatusMsg() == "Connecting..." then
		newStatusMsg("")	-- no longer show "connecting..."
	end
	
	local command, parameters = expandCommand(data)		-- after this instruction, command holds "ERROR" or "KARMA" or "LEFT"  etc.
	
	if command == "KARMA" then		-- new karma received		(parameter[1] is playername, parameter[2] is karma)
		-- check if user exists in table, otherwise add him:
		if users[parameters[1]] == nil then 
			createNewPlayer(parameters[1])
		end
		
		local newKarma = tonumber(parameters[2])
		
		users[parameters[1]]["karma"] = newKarma		-- set new karma for player
		
		if parameters[1] == getMyPlayerName() then	-- if I received new karma
			if parameters[3] ~= nil and users[parameters[3]] ~= nil then
				local dir = "+"
				local wasGood = true
				if myLastKarma - newKarma > 0 then dir = "-" end
				
				if (myLastKarma < 0 and dir == "-") or (myLastKarma > 0 and dir == "+") then
					wasGood = false		-- if it brought me closer to zero, it was good for me
				end
				
				addNewGivenKarma(parameters[3], dir, wasGood)		-- display that the user got Karma from another user
			end
			myLastKarma = newKarma
		end
		
	elseif command == "RESUME" then
		resumeGame()
	elseif command == "WINNER" then
		receiveNewWinner(parameters[1])
	elseif command == "ERROR" then
		if tonumber(parameters[1]) == ERROR_NAME_TAKEN then
			print ("disconnected...")
			connected = false
			newStatusMsg("Name already taken. Choose different playername!")
			resetMenu()
		else
			print(data);
		end
	elseif command == "LEFT" then			-- another client left the game
		if users[parameters[1]] ~= nil then 
			users[parameters[1]] = nil		-- remove client from playerlist
		end
	end
end

function receiveData()
	if not connected and tryingToConnect == false then return end
	repeat
		data, msg = udp:receive()
		
        if data then
			handleNewData(data)
		end
	until not data
end

function getIsConnected()
	return connected
end

function disconnect()
	if connected then 
		print("disconnecting from server")
		udp:send("LEAVE:")
	end
end